/* Zodiac - Md2Model.cpp
 */

#include "Md2Model.h"
#include "id_md2.h"
#include "posh.h"
#include <stdio.h>
#include <SDL_opengl.h>

static GLfloat blendVertices[MD2_MAX_VERTICES][3];
static GLfloat blendNormals[MD2_MAX_VERTICES][3];

Md2Model::Md2Model()
{
	m_numVertices = 0;
	m_numTexCoords = 0;
	m_numTriangles = 0;
	m_numSkins = 0;
	m_numFrames = 0;

	m_skins = NULL;
	m_texCoords = NULL;
	m_triangles = NULL;
	m_frames = NULL;
}

Md2Model::~Md2Model()
{
	if (m_skins != NULL) delete[] m_skins;
	if (m_texCoords != NULL) delete[] m_texCoords;
	if (m_triangles != NULL) delete[] m_triangles;

	if (m_frames != NULL)
	{
		for (int i = 0;i < m_numFrames; i++)
		{
			if (m_frames[i].vertices != NULL) delete[] m_frames[i].vertices;
		}
		delete[] m_frames;
	}
}

Md2Model* Md2Model::loadFromFile(const char *fileName)
{
	// Ignore garbage input
	if (!fileName || !fileName[0]) return NULL;

	md2_header header;
	FILE *f;
	char frame_buffer[MD2_MAX_FRAMESIZE];

	// Open the file, and make sure it's an MD2 model
	f = fopen(fileName, "rb");
	if (!f) return false;

	fread(&header, sizeof(md2_header), 1, f);

#ifndef POSH_LITTLE_ENDIAN
	// Swap the byte-ordering of the header on non little-endian machines
	header.magic = POSH_SwapU32(header.magic);
	header.version = POSH_SwapU32(header.version);
	header.skinWidth = POSH_SwapU32(header.skinWidth);
	header.skinHeight = POSH_SwapU32(header.skinHeight);
	header.frameSize = POSH_SwapU32(header.frameSize);
	header.numSkins = POSH_SwapU32(header.numSkins);
	header.numVertices = POSH_SwapU32(header.numVertices);
	header.numTexCoords = POSH_SwapU32(header.numTexCoords);
	header.numTriangles = POSH_SwapU32(header.numTriangles);
	header.numGlCommands = POSH_SwapU32(header.numGlCommands);
	header.numFrames = POSH_SwapU32(header.numFrames);
	header.offsetSkins = POSH_SwapU32(header.offsetSkins);
	header.offsetTexCoords = POSH_SwapU32(header.offsetTexCoords);
	header.offsetTriangles = POSH_SwapU32(header.offsetTriangles);
	header.offsetFrames = POSH_SwapU32(header.offsetFrames);
	header.offsetGlCommands = POSH_SwapU32(header.offsetGlCommands);
	header.offsetEnd = POSH_SwapU32(header.offsetEnd);
#endif

	if (header.magic != MD2_MAGIC_NUMBER || header.version != MD2_VERSION)
	{
		fclose(f);
		return NULL;
	}

	// Allocate memory to hold the model
	Md2Model* model = new Md2Model();
	model->m_numVertices = header.numVertices;
	model->m_numTexCoords = header.numTexCoords;
	model->m_numTriangles = header.numTriangles;
	model->m_numSkins = header.numSkins;
	model->m_numFrames = header.numFrames;

	model->m_texCoords = new Md2TexCoord[model->m_numTexCoords];
	model->m_triangles = new Md2Triangle[model->m_numTriangles];
	model->m_skins = new Md2SkinName[model->m_numSkins];
	model->m_frames = new Md2Frame[model->m_numFrames];

	for (int i = 0;i < model->m_numFrames; i++)
	{
		model->m_frames[i].vertices = new Md2Vertex[model->m_numVertices];
	}

	// Start loading data from the file.
	// Skins first
	fseek(f, header.offsetSkins, SEEK_SET);
	fread(model->m_skins, sizeof(Md2SkinName), model->m_numSkins, f);

	// Texture coordinates next
	fseek(f, header.offsetTexCoords, SEEK_SET);
	for (int i = 0;i < model->m_numTexCoords; i++)
	{
		// Convert the texture coordinates to normalized floats while
		// loading them
		md2_texcoord tc;
		fread(&tc, sizeof(tc), 1, f);

#ifndef POSH_LITTLE_ENDIAN
		// Convert the byte order of the shorts being loaded if not
		// on a little-endian machine
		tc.s = POSH_SwapU16(tc.s);
		tc.t = POSH_SwapU16(tc.t);
#endif

		model->m_texCoords[i].s = tc.s / (float)header.skinWidth;
		model->m_texCoords[i].t = tc.t / (float)header.skinHeight;
	}

	// Triangles can be loaded directly; their format on disk is the
	// same as the format in memory
	fseek(f, header.offsetTriangles, SEEK_SET);
	fread(model->m_triangles, sizeof(Md2Triangle), model->m_numTriangles, f);

#ifndef POSH_LITTLE_ENDIAN
	// Byte-swap the triangles too
	for (int i = 0;i < model->m_numTriangles; i++)
	{
		for (int v = 0;v < 3; v++)
		{
			model->m_triangles[i].vertexIndices[v] = POSH_SwapU16(model->m_triangles[i].vertexIndices[v]);
			model->m_triangles[i].texCoordIndices[v] = POSH_SwapU16(model->m_triangles[i].texCoordIndices[v]);
		}
	}
#endif

	// Last, load the frames of animation
	fseek(f, header.offsetFrames, SEEK_SET);
	for (int i = 0;i < model->m_numFrames; i++)
	{
		md2_frame *frame;

		// Read the current frame
		fread(frame_buffer, header.frameSize, 1, f);
		frame = (md2_frame*)frame_buffer;

#ifndef POSH_LITTLE_ENDIAN
		// Byte-swap the scale and translate vectors
#endif
		
		memcpy(model->m_frames[i].name, frame->name, 16);

		// Unpack the vertices for this frame
		for (int v = 0; v < model->m_numVertices; v++)
		{
			model->m_frames[i].vertices[v].x =
				frame->vertices[v].vertex[0] * frame->scale[0] + frame->translate[0];
			model->m_frames[i].vertices[v].y =
				frame->vertices[v].vertex[1] * frame->scale[1] + frame->translate[1];
			model->m_frames[i].vertices[v].z =
				frame->vertices[v].vertex[2] * frame->scale[2] + frame->translate[2];

			model->m_frames[i].vertices[v].normal = frame->vertices[v].lightNormalIndex;
		}

		// Calculate the bounding box for this frame
		model->m_frames[i].min[0] = frame->translate[0];
		model->m_frames[i].min[1] = frame->translate[1];
		model->m_frames[i].min[2] = frame->translate[2];
		model->m_frames[i].max[0] = frame->translate[0] + (frame->scale[0] * 255.0f);
		model->m_frames[i].max[1] = frame->translate[1] + (frame->scale[1] * 255.0f);
		model->m_frames[i].max[2] = frame->translate[2] + (frame->scale[2] * 255.0f);
	}

	// Done
	fclose(f);
	return model;
}

const Md2SkinName *Md2Model::getSkin(int index) const
{
	if (index >= 0 && index < m_numSkins)
	{
		return &m_skins[index];
	}
	return NULL;
}

const Md2Frame *Md2Model::getFrame(int index) const
{
	if (index >= 0 && index < m_numFrames)
	{
		return &m_frames[index];
	}
	return NULL;
}

void Md2Model::blendFrames(int frame1, int frame2, float lerp) const
{
	int i;
	const Md2Frame *from, *to;

	from = getFrame(frame1);
	to = getFrame(frame2);

	if(lerp <= 0)
	{
		// copy the vertices in frame 'from' over
		for(i = 0;i < m_numVertices;i++)
		{
			blendVertices[i][0] = from->vertices[i].x;
			blendVertices[i][1] = from->vertices[i].y;
			blendVertices[i][2] = from->vertices[i].z;

			blendNormals[i][0] = kMd2Normals[from->vertices[i].normal][0];
			blendNormals[i][1] = kMd2Normals[from->vertices[i].normal][1];
			blendNormals[i][2] = kMd2Normals[from->vertices[i].normal][2];
		}
	} else if(lerp >= 1.0f)
	{
		// copy the vertices in frame 'to'
		for(i = 0;i < m_numVertices;i++)
		{
			blendVertices[i][0] = to->vertices[i].x;
			blendVertices[i][1] = to->vertices[i].y;
			blendVertices[i][2] = to->vertices[i].z;

			blendNormals[i][0] = kMd2Normals[to->vertices[i].normal][0];
			blendNormals[i][1] = kMd2Normals[to->vertices[i].normal][1];
			blendNormals[i][2] = kMd2Normals[to->vertices[i].normal][2];
		}
	} else
	{
		// mix the vertices
		for(i = 0;i < m_numVertices;i++)
		{
			blendVertices[i][0] = from->vertices[i].x +
				(to->vertices[i].x - from->vertices[i].x) * lerp;
			blendVertices[i][1] = from->vertices[i].y +
				(to->vertices[i].y - from->vertices[i].y) * lerp;
			blendVertices[i][2] = from->vertices[i].z +
				(to->vertices[i].z - from->vertices[i].z) * lerp;

			blendNormals[i][0] = kMd2Normals[from->vertices[i].normal][0] +
				(kMd2Normals[to->vertices[i].normal][0] - kMd2Normals[from->vertices[i].normal][0]) * lerp;
			blendNormals[i][0] = kMd2Normals[from->vertices[i].normal][1] +
				(kMd2Normals[to->vertices[i].normal][1] - kMd2Normals[from->vertices[i].normal][1]) * lerp;
			blendNormals[i][0] = kMd2Normals[from->vertices[i].normal][2] +
				(kMd2Normals[to->vertices[i].normal][2] - kMd2Normals[from->vertices[i].normal][2]) * lerp;
		}
	}
}

void Md2Model::drawFrame(int frame) const
{
	drawBlendedFrames(frame, frame, 0);
}

void Md2Model::drawBlendedFrames(int frame1, int frame2, float lerp) const
{
	if (frame1 < 0 || frame1 >= m_numFrames) return;
	if (frame2 < 0 || frame2 >= m_numFrames) return;

	blendFrames(frame1, frame2, lerp);

	int i;
	Md2Triangle *tri;

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	glVertexPointer(3, GL_FLOAT, 0, blendVertices);
	glNormalPointer(GL_FLOAT, 0, blendNormals);

	glBegin(GL_TRIANGLES);
	{
		for(i = 0;i < m_numTriangles;i++)
		{
			tri = &m_triangles[i];

			glTexCoord2fv((const GLfloat*)&(m_texCoords[tri->texCoordIndices[0]]));
			glArrayElement(tri->vertexIndices[0]);
			
			glTexCoord2fv((const GLfloat*)&(m_texCoords[tri->texCoordIndices[1]]));
			glArrayElement(tri->vertexIndices[1]);

			glTexCoord2fv((const GLfloat*)&(m_texCoords[tri->texCoordIndices[2]]));
			glArrayElement(tri->vertexIndices[2]);
		}
	}
	glEnd();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
}
