/* Zodiac - ModelViewer.cpp
 */

#include "ModelViewer.h"
#include <SDL.h>
#include <SDL_opengl.h>
using namespace std;

// Based on visual observation, it appears to be that Egoboo's modeled were
// designed to play back at 5 frames per second.
static const float kAnimateSpeed = 1.0f / 5.0f; 

ModelViewer::ModelViewer()
{
	m_background = NULL;
	m_model = NULL;
	m_modelTxr = NULL;
	ui = NULL;

	m_modelPage = 0;
}

ModelViewer::~ModelViewer()
{
	delete m_background;
	delete m_model;
	delete m_modelTxr;
	delete ui;
}

bool ModelViewer::loadResources()
{
	bool result = true;

	m_background = Texture::loadFromFile("background.bmp");
	if (!m_background)
	{
		m_log.printf("TestMenu: Could not load background.bmp!\n");
		result = false;
	}

	m_modelAngle = 0;
	m_model = NULL;
	m_modelTxr = NULL;

	m_fileList = sys_listFiles("./objects/*.obj");
	m_numModels = m_fileList.size();

	Font *f = Font::loadFromFile("tahoma.ttf", 16);
	if (!f)
	{
		m_log.printf("ModelViewer: Could not load Papyrus.ttf!\n");
		result = false;
	}
	ui = new UiContext();
	ui->setFont(f);

	return result;
}

bool ModelViewer::doFrame(double deltaTime)
{
	bool rval = doMenu();
	doModel(deltaTime);
	return rval;
}

bool ModelViewer::doMenu()
{
	int buttonID = 1;

	ui->beginFrame(0);

	if (m_background && !m_model)
	{
		int x, y, width, height;
		width = m_displayWidth;
		height = m_displayHeight;

		x = width - m_background->sourceWidth();
		y = height - m_background->sourceHeight();
		ui->drawImage(m_background, UiRect(x, y, 0, 0));
	}

	Font *font = ui->font();
	font->drawTextFormatted(20, 20, "%d models available", m_numModels);

	list<string>::iterator i, end;
	UiRect rect(20, 60, 125, 25);
	for (i = m_fileList.begin(), end = m_fileList.end();
		i != end;
		++i, ++buttonID)
	{
		if (ui->doButton(buttonID, (*i).c_str(), rect))
		{
			loadModel((*i).c_str());
		}

		rect.y += rect.height;
	}

	rect.width = 100;
	rect.x = m_displayWidth - 20 - rect.width * 2;
	rect.y = m_displayHeight - 10 - rect.height;
	if (ui->doButton(buttonID, "Clear Model", rect))
	{
		delete m_model;
		delete m_modelTxr;
		m_model = NULL;
		m_modelTxr = NULL;
	}

	rect.x += rect.width + 10;
	buttonID++;
	bool rVal = !ui->doButton(buttonID, "Quit", rect);

#ifndef NDEBUG
	if (ui->mouseBeingDragged())
	{
		font->drawText(20, m_displayHeight - 120, "Mouse being dragged.");
		UiPoint distance = ui->mouseDragDistance();
		font->drawTextFormatted(20, m_displayHeight - 100, "Mouse drag distance: %d, %d", distance.x, distance.y);
	}
	font->drawTextFormatted(20, m_displayHeight - 80, "Hot control is id %d", ui->hotControl());
	font->drawTextFormatted(20, m_displayHeight - 60, "Active control is id %d", ui->activeControl());
#endif

	// Need to update the model's drawing angle before ui->endFrame is called, because that's when the
	// last frame position gets updated.
	m_modelAngle += ui->mouseDragDistance().x * 0.6f;

	// Let the viewer know what frame we're on
	if (m_model != NULL)
	{
		font->drawTextFormatted(20, m_displayHeight - 40, "Currently on frame %d, named %s", m_modelFrame, m_model->getFrame(m_modelFrame)->name);
	}

	ui->endFrame();

	return rVal;
}

void ModelViewer::doModel(double deltaTime)
{
	if (m_model && m_modelTxr)
	{
		// Advance the model's animation
		m_modelLerp += (float)deltaTime / kAnimateSpeed;
		if (m_modelLerp >= 1.0f)
		{
			m_modelFrame = m_modelNextFrame;
			m_modelNextFrame++;

			// Roll back to the first frame if we run out of them
			if (m_modelNextFrame >= m_model->numFrames())
			{
				m_modelNextFrame = 0;
			}

			m_modelLerp -= 1.0f;
		}

		float width, height;
		width =(float)m_displayWidth;
		height = (float)m_displayHeight;

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(60.0, width/height, 1, 2048);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(20, -20, -75);
		glRotatef(-90, 1, 0, 0);
		glRotatef(m_modelAngle, 0, 0, 1);

		glEnable(GL_TEXTURE_2D);
		glEnable(GL_DEPTH_TEST);

		glBindTexture(GL_TEXTURE_2D, m_modelTxr->binding());
		m_model->drawBlendedFrames(m_modelFrame, m_modelNextFrame, m_modelLerp);
	}
}

int ModelViewer::run(int width, int height, char *logFile)
{
	double frameTime, lastFrameTime;

	m_displayWidth = width;
	m_displayHeight = height;

	if (logFile != NULL)
	{
		m_log.open(logFile);
		m_log.printf("ModelViewer starting up.\nUsing a %d x %d window.\n", width, height);
	}

	if (!loadResources())
	{
		return -1;
	}

	// Initialize the system timers.  High performance timing, whoo!
	sys_initialize();
	frameTime = lastFrameTime = sys_getTime();
	bool bAlive = true;
	while (bAlive)
	{
		SDL_Event evt;
		while (SDL_PollEvent(&evt))
		{
			if (evt.type == SDL_QUIT)
			{
				bAlive = false;
			}

			ui->handleSDLEvent(&evt);
		}

		if (!bAlive) break;

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		bAlive = doFrame(frameTime - lastFrameTime);

		SDL_GL_SwapBuffers();

		// I don't need to suck up every tiny bit of CPU time at the moment.
		SDL_Delay(1);
		lastFrameTime = frameTime;
		frameTime = sys_getTime();
	}

	return 0;
}

void ModelViewer::loadModel(string model)
{
	delete m_model;
	delete m_modelTxr;

	string name = "objects/" + model + "/tris.md2";
	m_log.printf("Loading model %s: ", name.c_str());
	m_model = Md2Model::loadFromFile(name.c_str());
	if (m_model) {
		m_log.printf("Success.\n");
	} else {
		m_log.printf("Failed\n");
	}

	name = "objects/" + model + "/tris0.bmp";
	m_log.printf("Loading texture %s: ", name.c_str());
	m_modelTxr = Texture::loadFromFile(name.c_str());
	if (m_modelTxr) {
		m_log.printf("Success.\n");
	} else {
		m_log.printf("Failed\n");
	}

	// Reset the model viewing angle so that the model's looking
	// relatively forward when it gets loaded.  For aesthetics, ya know.
	m_modelAngle = -100;

	// And start the animations out fresh
	m_modelFrame = 0;
	m_modelNextFrame = 1;
	m_modelLerp = 0;
}
