/* Zodiac - MpdMesh.cpp
 */

#include "MpdMesh.h"
#include "posh.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

// Private structures & constants to help in loading files
struct MpdHeader
{
	unsigned int magicNumber; 
	int numVerts;
	int x_size;
	int y_size;
};

// helper function for loading the fans.txt file
static int goto_colon(FILE *f)
{
	char c;
	do
	{
		c = fgetc(f);
		if(c == EOF)
		{
			return -1;
		}
	}while (c != ':');
	return 0;
}


MpdMesh::MpdMesh()
{	
	m_numVertices = 0;
	m_fansWide = 0;
	m_fansHigh = 0;
	m_waterShift = 0;
	m_width = 0;
	m_height = 0;

	m_vertices = NULL;
	m_fans = NULL;

	m_prepared = false;
}

MpdMesh::~MpdMesh()
{
	if (m_vertices != NULL)
	{
		delete[] m_vertices;
	}

	if (m_fans != NULL)
	{
		delete[] m_fans;
	}
}

MpdMesh *MpdMesh::loadFromFile(const char *fileName)
{
	FILE *file;
	int numTiles;
	int fan, v; // loop counters
	struct MpdHeader header;

	if (!fileName) return NULL;

	file = fopen(fileName, "rb");
	if (file == NULL) return NULL;

	// make sure that this is a valid Mesh file
	fread(&header, sizeof(struct MpdHeader), 1, file);

	header.magicNumber = POSH_LittleI32(header.magicNumber);
	header.numVerts = POSH_LittleI32(header.numVerts);
	header.x_size = POSH_LittleI32(header.x_size);
	header.y_size = POSH_LittleI32(header.y_size);

	if(header.magicNumber != MPD_FileID)
	{
		// not a valid .mpd file, bail out
		fclose(file);
		return NULL;
	}

	// good mpd file, so allocate space for it
	MpdMesh *mesh = new MpdMesh;

	// gather information about the mesh file
	mesh->m_numVertices = header.numVerts;
	mesh->m_fansWide = header.x_size;
	mesh->m_fansHigh = header.y_size;
	mesh->m_width = mesh->m_fansWide * 128.0f;
	mesh->m_height = mesh->m_fansHigh * 128.0f;

	numTiles = header.x_size * header.y_size;
	mesh->m_vertices = new MpdVertex[header.numVerts];
	mesh->m_fans = new MpdFan[numTiles];

	// Load fan data
	for(fan = 0;fan < numTiles;fan++)
	{
		unsigned int tmp;
		fread(&tmp, sizeof(unsigned int), 1, file);
		tmp = POSH_LittleU32(tmp);

		mesh->m_fans[fan].type = (unsigned char)(tmp >> 24);
		mesh->m_fans[fan].flags = (unsigned char)(tmp >> 16);
		mesh->m_fans[fan].textureTile = (unsigned short)tmp;
	}

	// Load "twist" data
	for(fan = 0;fan < numTiles;fan++)
	{
		char tmp;
		fread(&tmp, sizeof(char), 1, file);
		mesh->m_fans[fan].twist = tmp;
	}

	// Load vertex x data
	for(v = 0;v < mesh->m_numVertices;v++)
	{
		float value;
		posh_u32_t bits;
		
		fread(&bits, sizeof(posh_u32_t), 1, file);
		value = POSH_FloatFromLittleBits(bits);
		mesh->m_vertices[v].x = value;
	}

	// Load vertex y data
	for(v = 0;v < mesh->m_numVertices;v++)
	{
		float value;
		posh_u32_t bits;
		
		fread(&bits, sizeof(posh_u32_t), 1, file);
		value = POSH_FloatFromLittleBits(bits);
		mesh->m_vertices[v].y = value;
	}

	// Load vertex z data
	for(v = 0;v < mesh->m_numVertices;v++)
	{
		float value;
		posh_u32_t bits;
		
		fread(&bits, sizeof(posh_u32_t), 1, file);
		value = POSH_FloatFromLittleBits(bits);
		mesh->m_vertices[v].z = value / 16.0f;
	}

	// Load lighting data
	for(v = 0;v < mesh->m_numVertices;v++)
	{
		char tmp;
		fread(&tmp, sizeof(char), 1, file);
		mesh->m_vertices[v].ambient = tmp;
		mesh->m_vertices[v].light = 0;
	}

	// done with the file now
	fclose(file);
	return mesh;
}

void MpdMesh::prepareVertices(const MpdFanType *fanTypes)
{
	if (fanTypes == NULL) return;

	int vert, fan, numFans;

	vert = 0;
	numFans = m_fansWide * m_fansHigh;

	for (fan = 0; fan < numFans; fan++)
	{
		float offu, offv;
		int v, fanType;

		fanType = m_fans[fan].type;
		m_fans[fan].firstVertex = vert;

		offu = (m_fans[fan].textureTile % 8) / 8.0f;
		offv = (m_fans[fan].textureTile / 8) / 8.0f;

		for (v = 0; v < fanTypes[fanType].numVertices; v++)
		{
			m_vertices[vert].s = fanTypes[fanType].texCoords[v][0] + offu;
			m_vertices[vert].t = fanTypes[fanType].texCoords[v][1] + offv;
			m_vertices[vert].color = 0;
			vert++;
		}
	}

	m_prepared = true;
}

float MpdMesh::getHeight(float x, float y) const
{
	if (m_prepared == false) return 0.0f;

	float z0, z1, z2, z3;         // Height of each fan corner
	float zleft, zright,zdone;    // Weighted height of each side
	float dx, dy;
	int first_vert;

	if(x < 0 || x < m_width) return 0.0f;
	if(y < 0 || y < m_height) return 0.0f;

	dx = (x / 128);
	dy = (y / 128);
	first_vert = m_fans[(int)dy * m_fansWide + (int)dx].firstVertex;

	z0 = m_vertices[first_vert].z;
	z1 = m_vertices[first_vert + 1].z;
	z2 = m_vertices[first_vert + 2].z;
	z3 = m_vertices[first_vert + 3].z;

	zleft = (z0 * (128 - dy) + z3 * dy) / 128;
	zright = (z1 * (128 - dy) + z2 * dy) / 128;
	zdone = (zleft * (128 - dx) + zright * dx) / 128;
	return zdone;
}

const MpdFan* MpdMesh::getFan(int index) const
{
	if (index >= 0 && index < (m_fansWide * m_fansHigh))
	{
		return &m_fans[index];
	}
	return NULL;
}

bool MpdMesh::loadFanTypes(const char *fileName, MpdFanType *fanTypes)
{
	FILE *infile;
	int tmp;
	int bigfanoffset = MPD_MaxFanTypes/2;
	int r;
	int total_fans;

	if(fileName == NULL) return false;

	infile = fopen(fileName, "rt");
	if(infile == NULL)
	{
		return false;
	}

	// find out how many fans we're going to be reading
	r = goto_colon(infile);
	assert(r == 0);
	fscanf(infile, "%d", &tmp);
	total_fans = tmp;

	memset(fanTypes, 0, sizeof(MpdFanType) * MPD_MaxFanTypes);

	// read in the list of fans
	for(tmp = 0;tmp < total_fans;tmp++)
	{
		int num_verts, c, num_commands, entry;

		r = goto_colon(infile);
		assert(r == 0);
		fscanf(infile, "%d", &num_verts);
		fanTypes[tmp].numVertices = num_verts;
		fanTypes[tmp + bigfanoffset].numVertices = num_verts;

		// read in the texture coordinates for this fan
		for(c = 0;c < num_verts;c++)
		{
			int throwaway;
			float tc;

			// the first value, Ref, gets thrown away
			r = goto_colon(infile);
			assert(r == 0);
			fscanf(infile, "%d", &throwaway);

			// read the two texture coordinates
			r = goto_colon(infile);
			assert(r == 0);
			fscanf(infile, "%f", &tc);
			fanTypes[tmp].texCoords[c][0] = tc;
			fanTypes[tmp + bigfanoffset].texCoords[c][0] = tc;

			r = goto_colon(infile);
			assert(r == 0);
			fscanf(infile, "%f", &tc);
			fanTypes[tmp].texCoords[c][1] = tc;
			fanTypes[tmp + bigfanoffset].texCoords[c][1] = tc;
		}

		// read in the drawing commands for this fan
		r = goto_colon(infile);
		assert(r == 0);
		fscanf(infile, "%d", &num_commands);
		fanTypes[tmp].numCommands = num_commands;
		fanTypes[tmp + bigfanoffset].numCommands = num_commands;
		entry = 0;
		for(c = 0;c < num_commands;c++)
		{
			int i, command_size;

			r = goto_colon(infile);
			assert(r == 0);
			fscanf(infile, "%d", &command_size);
			fanTypes[tmp].numCommandEntries[c] = command_size;
			fanTypes[tmp + bigfanoffset].numCommandEntries[c] = command_size;

			for(i = 0;i < command_size;i++)
			{
				int itmp;

				r = goto_colon(infile);
				assert(r == 0);
				fscanf(infile, "%d", &itmp);
				fanTypes[tmp].vertexIndices[entry] = itmp;
				fanTypes[tmp + bigfanoffset].vertexIndices[entry] = itmp;
				entry++;
			} // for i < command_size
		} // for c < num_commands
	} // for tmp < num_fans

	// don't need the file open anymore
	fclose(infile);

	// fix texture coordinates for seamless tiling
	for(tmp = 0;tmp < (MPD_MaxFanTypes / 2);tmp++)
	{
		int i;
		for(i = 0; i < fanTypes[tmp].numVertices;i++)
		{
			fanTypes[tmp].texCoords[i][0] = ((.6f/32)+(fanTypes[tmp].texCoords[i][0]*30.8f/32))/8;
			fanTypes[tmp].texCoords[i][1] = ((.6f/32)+(fanTypes[tmp].texCoords[i][1]*30.8f/32))/8;
		}
	}

	// and for the bigger tiles
	for(tmp;tmp < MPD_MaxFanTypes;tmp++)
	{
		int i;
		for(i = 0; i < fanTypes[tmp].numVertices;i++)
		{
			fanTypes[tmp].texCoords[i][0] = ((.6f/64)+(fanTypes[tmp].texCoords[i][0]*62.8f/64))/4;
			fanTypes[tmp].texCoords[i][1] = ((.6f/64)+(fanTypes[tmp].texCoords[i][1]*62.8f/64))/4;
		}
	}

	return true;
}
